#import <AR/gsub_lite.h>
#import "ARMarkerDetector.h"


NSMutableDictionary *markerObjectForPath = nil;
NSMutableDictionary *pathRetainCounts = nil;

@implementation ARMarkerDetector
+(BOOL)isSafe
{
	return NO;
}

+(BOOL)allowsSubpatchesWithIdentifier:(id)identifier
{
	return NO;
}

+(int)executionModeWithIdentifier:(id)identifier
{
	return 0;
}

+(int)timeModeWithIdentifier:(id)identifier
{
	return 0;
}

-(id)initWithIdentifier:(id)identifier
{
	if(self = [super initWithIdentifier:identifier])
	{
		[[self userInfo] setObject:@"Kineme AR Marker Detector" forKey:@"name"];
		
		[inputThreshold setMaxIndexValue:255];
		[inputThreshold setIndexValue:100];
		
		[inputPatternMillimeters setDoubleValue:80];
		
		patternPaths = [NSMutableArray new];
		if (! markerObjectForPath)
			markerObjectForPath = [NSMutableDictionary new];
		if (! pathRetainCounts)
			pathRetainCounts = [NSMutableDictionary new];
	}
	return self;
}

-(BOOL)setup:(QCOpenGLContext*)context
{
	return YES;
}

-(void)cleanup:(QCOpenGLContext*)context
{
}

-(void)enable:(QCOpenGLContext*)context
{
}

-(void)disable:(QCOpenGLContext*)context
{
}


-(void)updatePatternMillimeters
{
	double patternMillimeters = [inputPatternMillimeters doubleValue];
	
	for (NSString *path in patternPaths)
	{
		ObjectData_T *mo = [[markerObjectForPath objectForKey:path] pointerValue];
		mo->marker_width = patternMillimeters;
	}
}

-(void)updatePatternFiles
{
	for (NSString *path in patternPaths)
	{
		int retainCount = [[pathRetainCounts objectForKey:path] intValue] - 1;
		if (retainCount == 0)
		{
			ObjectData_T *mo = [[markerObjectForPath objectForKey:path] pointerValue];
			arFreePatt(mo->id);
			free(mo);
			[markerObjectForPath removeObjectForKey:path];
			[pathRetainCounts removeObjectForKey:path];
		}
		else
		{
			[pathRetainCounts setObject:[NSNumber numberWithInt:retainCount] forKey:path];
		}
	}
	[patternPaths removeAllObjects];
	
	NSUInteger i = 0;
	for (id patternFile in [[inputPatternFiles structureValue] _list])
	{
		if ([patternFile isKindOfClass:[NSString class]])
		{
			NSString *path = ([(NSString *)patternFile length] > 0 ? 
							  KIExpandPath(self, (NSString *)patternFile) : 
							  nil);
			
			int retainCount;
			if (! [markerObjectForPath objectForKey:path])
			{
				ObjectData_T *mo = (ObjectData_T *)malloc(sizeof(ObjectData_T));
				mo->visible = 0;
				strncpy(mo->name, [[path lastPathComponent] UTF8String], 256);
				mo->id = arLoadPatt([path UTF8String]);
				if (mo->id < 0)
					NSLog(NSLocalizedString(@"Couldn't load pattern file %@", @""), path);
				mo->marker_width = [inputPatternMillimeters doubleValue];
				mo->marker_center[0] = 0;
				mo->marker_center[1] = 0;
				
				[markerObjectForPath setObject:[NSValue valueWithPointer:mo] forKey:path];
				retainCount = 1;
			}
			else
			{
				retainCount = [[pathRetainCounts objectForKey:path] intValue] + 1;
			}
			
			[pathRetainCounts setObject:[NSNumber numberWithInt:retainCount] forKey:path];
			[patternPaths addObject:path];
		}
		++i;
	}
}

-(void)updatePatternStructureForContext:(QCOpenGLContext *)context
{
	if ([patternPaths count] == 0)
	{
		[outputMarkerStructure setStructureValue:nil];
		return;
	}
	
	QCStructure *structure = nil;
	
	QCImage *image = [inputImage imageValue];
	CGColorSpaceRef colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
	QCImagePixelBuffer *pb = [image createPixelBufferForManager:[context imageManager] 
													 withFormat:[QCPixelFormat pixelFormatARGB8]
														 bounds:[image bounds]
													 colorSpace:colorSpace
														options:nil];
	[pb beginUpdatePixels:YES colorSpace:colorSpace];
	
	
	arParamChangeSize(&cameraParam, [image bounds].size.width, [image bounds].size.height, &arParam);
	arInitCparam(&arParam);
	
	ARMarkerInfo *marker_info;
	int marker_num;
	ARUint8 *gARTImage = [pb baseAddress];
	
	if (arDetectMarkerLite(gARTImage, [inputThreshold indexValue], &marker_info, &marker_num) < 0) {  // arDetectMarker does not work correctly after arFreePatt
		NSLog(NSLocalizedString(@"Couldn't do marker detection", @""));
		goto END_DETECT;
	}
	
	GFList *structureList = [[GFList alloc] initWithCapacity:[patternPaths count]];
	NSUInteger i = 0;
	for (NSString *path in patternPaths)
	{
		ObjectData_T *markerObject = [[markerObjectForPath objectForKey:path] pointerValue];
		
		int k = -1;
		for (int j = 0; j < marker_num; j++ ) {
			if( markerObject->id == marker_info[j].id ) {
				if( k == -1 ) k = j;
				else if( marker_info[k].cf < marker_info[j].cf ) k = j;
			}
		}
		
		QCStructure *objectStructure = [[QCStructure alloc] init];
		[objectStructure setMember:[NSNumber numberWithBool:(markerObject->id < 0)] forKey:@"Error"];
		
		markerObject->visible = (k != -1);
		[objectStructure setMember:[NSNumber numberWithBool:markerObject->visible] forKey:@"Found"];
		
		if (markerObject->visible)
		{
			arGetTransMat(&marker_info[k], markerObject->marker_center, markerObject->marker_width, markerObject->trans);
			
			GLdouble modelMat[16];
			arglCameraViewRH(markerObject->trans, modelMat, 0.025 /* scale */);
			
			[objectStructure setMember:[NSNumber numberWithDouble: modelMat[12]] forKey:@"X"];
			[objectStructure setMember:[NSNumber numberWithDouble: modelMat[13]] forKey:@"Y"];
			[objectStructure setMember:[NSNumber numberWithDouble: 1.5*modelMat[14]] forKey:@"Z"];
			
			// http://www.codeguru.com/forum/archive/index.php/t-329530.html
			CGFloat rx = atan2(modelMat[9], modelMat[10]) * 180/3.14;
			CGFloat ry = -asin(modelMat[8]) * 180/3.14;
			CGFloat rz = atan2(modelMat[4], modelMat[0]) * 180/3.14;
			[objectStructure setMember:[NSNumber numberWithDouble:-rx] forKey:@"RX"];
			[objectStructure setMember:[NSNumber numberWithDouble:-ry] forKey:@"RY"];
			[objectStructure setMember:[NSNumber numberWithDouble:-rz] forKey:@"RZ"];
			
			NSMutableArray *matrixArray = [[NSMutableArray alloc] init];
			for (int j = 0; j < 16; ++j)
				[matrixArray addObject:[NSNumber numberWithDouble:modelMat[j]]];
			QCStructure *matrixStructure = [[QCStructure alloc] initWithArray:matrixArray];
			[objectStructure setMember:matrixStructure forKey:@"Matrix"];
			[matrixStructure release];
			[matrixArray release];
		}
		
		[structureList insertObject:objectStructure atIndex:i forKey:[NSString stringWithFormat:@"%s", markerObject->name]];
		[objectStructure release];
		++i;
	}
	
	structure = [[QCStructure alloc] init];
	[structure _setList: structureList];
	[structureList release];
	
END_DETECT:
	[pb endUpdatePixels];
	[pb release];
	CGColorSpaceRelease(colorSpace);
	
	[outputMarkerStructure setStructureValue:structure];
	[structure release];
}


-(BOOL)execute:(QCOpenGLContext*)context time:(double)time arguments:(NSDictionary*)arguments
{
	if ([inputCameraCalibrationFile wasUpdated])
	{
		NSString *path = ([[inputCameraCalibrationFile stringValue] length] > 0 ? 
						  KIExpandPath(self, [inputCameraCalibrationFile stringValue]) : 
						  [[NSBundle bundleForClass:[self class]] pathForResource:@"camera_para" ofType:@"dat"]);
		
		if (arParamLoad([path UTF8String], 1, &cameraParam) < 0)
			NSLog(NSLocalizedString(@"Couldn't load camera parameters file %@", @""), path);
	}
	
	if ([inputPatternMillimeters wasUpdated] && ! [inputPatternFiles wasUpdated])
	{
		[self updatePatternMillimeters];
	}
	
	if ([inputPatternFiles wasUpdated])
	{
		[self updatePatternFiles];
	}
	
	[self updatePatternStructureForContext:context];
	
	return YES;
}

@end
